program Touch32;
{$APPTYPE CONSOLE}
{ $DEFINE PAUSE}
uses
  KOL, Windows;
const Info =
'Touch32                             v.0.96 for Win32, 2009-002-10, A. Kubaszek'^M^J+
^M^J+
'Sets the date and time that a file (or directory in WinXP) was created,'^M^J+
'last accessed or last modified.'^M^J+
^M^J+
'Syntax: Touch32 File [ /[w][a][c][[f]|[d]] ] [ RefFile ] [ DateTime ]'^M^J+
^M^J+
'A File must exist. Use your local format for DateTime parameter (local time is'^M^J+
'converted to UTC). If you do not specify DateTime, the current time is used.'^M^J+
'Switches are set of characters:'^M^J+
'  w - update the Last Write Time (default)'^M^J+
'  a - update the Last Access Time'^M^J+
'  c - update the Creation Time'^M^J+
'  d - update the Last Write H:M, to show version (D2..D10) of DCU _File_'^M^J+
'  f - use the corresponding timestamps from the file RefFile.'^M^J+
'      RefFile parameter must be defined in this case.'^M^J+
'      Switches w,a,c can be also defined.'^M^J+
^M^J+
'Examples:                  (errorlevel = 0 or 2..10(/d) or 1013 if error occur)'^M^J+
'  Touch32 MyFile.txt /wa'^M^J+
'  Touch32 "My file.txt" /f "Ref file.txt" '^M^J+
'  Touch32 Unit.DCU /d'^M^J+
'  Touch32 "My file.txt" ';

type
  TSwitches = set of (w,a,c,f,d);

function FileTime2SysTime( const FT: TFileTime ) : TSystemTime;
begin FileTimeToSystemTime( FT, Result ); end;

function SysTime2DateTime(const SysTm :TSystemTime ): TDateTime;
begin SystemTime2DateTime( SysTm, Result ); end;

function FileTime2Str( const FT: TFileTime ) :string;
begin
  Result:=DateTime2StrShort(SysTime2DateTime(FileTime2SysTime( FT )));
end;

function ParamStrEnd( Idx: Integer ): String;
  function SkipParam(P: PChar): PChar;
  begin
    P := SkipSpaces( P );
    while P[0] > ' ' do
      if P[0] = '"' then
      begin
        Inc(P);
        while (P[0] <> #0) and (P[0] <> '"') do
          Inc(P);
        if P[0] <> #0 then Inc(P);
      end
        else
        Inc(P);
    Result := P;
  end;

var
  P, P1: PChar;
begin
  if Idx = 0 then
    Result := GetCommandLine
  else
  begin
    P := GetCommandLine;
    repeat
      P := SkipSpaces( P );
      P1 := P;
      P := SkipParam(P);
      if Idx = 0 then Break;
      Dec(Idx);
    until (Idx < 0) or (P = P1);
    Result := P1;
    // Result := Copy( P1, 1, P - P1 );
    if Length( Result ) >= 2 then
      if (Result[ 1 ] = '"') and (Result[ Length( Result ) ] = '"') then
        Result := Copy( Result, 2, Length( Result ) - 2 );
  end;
end;


///////przeniesione z KOL.pas
// + ECM

//[function SystemTime2DateTime]
//AK - korekta gdy dd lub mm = 0

function SystemTime2DateTime(const SystemTime : TSystemTime; var DateTime : TDateTime ) : Boolean;
var I : Integer;
    _Day, _Month : Integer;
    DayTable: PDayTable;
begin
  Result := False;
  DateTime := 0.0;
  DayTable := @MonthDays[IsLeapYear(SystemTime.wYear)];

  with SystemTime do
  if {(wYear >= 0) !always true! and} (wYear <= 9999) and
    {(wMonth >= 1) and !otherwise can not convert time only!}
    (wMonth <= 12) and
    {(wDay >= 1) and !otherwise can not convert time only!}
    (wDay <= DayTable^[wMonth]) and                                      //
    (wHour < 24) and (wMinute < 60) and (wSecond < 60) and (wMilliSeconds < 1000) then   //
  begin
    _Day := wDay;
    _Month := wMonth;
    if _Day=0 then _Day:=1;
    if _Month=0 then _Month:=1;

    for I := 1 to _Month - 1 do
        Inc(_Day, DayTable^[I]);
    I := wYear - 1;
    //--------------- by Vadim Petrov ------++
    if I<0 then i := 0;                     //
    //--------------------------------------++
    DateTime := I * 365 + I div 4 - I div 100 + I div 400 + _Day
             + (wHour * 3600000 + wMinute * 60000 + wSecond * 1000 + wMilliSeconds) / MSecsPerDay;
    Result := True;
  end;
end;



//[function Str2DateTimeFmt]
function Str2DateTimeFmt( const sFmtStr, sStr: KOLString ): TDateTime;
var h12, hAM: Boolean;
    FmtStr, S: PKOLChar;
    sS:KOLString;

  function GetNum( var S: PKOLChar; NChars: Integer ): Integer;
  begin
    Result := 0;
    while (S^ <> #0) and (NChars <> 0) do
    begin
      Dec( NChars );
      {$IFDEF UNICODE_CTRLS}
      if (S^ >= '0') and (S^ <= '9') then
      {$ELSE}
      if S^ in ['0'..'9'] then
      {$ENDIF}
      begin
        Result := Result * 10 + Ord(S^) - Ord('0');
        Inc( S );
      end
      else
        break;
    end;
  end;

  function GetYear( var S: PKOLChar; NChars: Integer ): Integer;
  var STNow: TSystemTime;
      OldDate: Boolean;
  begin
    Result := GetNum( S, NChars );
    GetSystemTime( STNow );
    OldDate := Result < 50;
    Result := Result + STNow.wYear - STNow.wYear mod 100;
    if OldDate then Dec( Result, 100 );
  end;

  function GetMonth( const fmt: KOLString; var S: PKOLChar ): Integer;
  var SD: TSystemTime;
      M: Integer;
      C, MonthStr: KOLString;
  begin
    GetSystemTime( SD );
    for M := 1 to 12 do
    begin
      SD.wMonth := M;
      C := SystemDate2Str( SD, LOCALE_USER_DEFAULT, dfLongDate, PKOLChar( fmt + '/dd/yyyy/' ) );
      MonthStr := Parse( C, '/' );
      if AnsiCompareStrNoCase( MonthStr, Copy( S, 1, Length( MonthStr ) ) ) = 0 then
      begin
        Result := M;
        Inc( S, Length( MonthStr ) );
        Exit;
      end;
    end;
    Result := 1;
  end;

  procedure SkipDayOfWeek( const fmt: KOLString; var S: PKOLChar );
  var SD: TSystemTime;
      Dt: TDateTime;
      D: Integer;
      C, DayWeekStr: KOLString;
  begin
    GetSystemTime( SD );
    SystemTime2DateTime( SD, Dt );
    Dt := Dt - SD.wDayOfWeek;
    for D := 0 to 6 do
    begin
      DateTime2SystemTime( Dt, SD );
      C := SystemDate2Str( SD, LOCALE_USER_DEFAULT, dfLongDate, PKOLChar( fmt + '/MM/yyyy/' ) );
      DayWeekStr := Parse( C, '/' );
      if AnsiCompareStrNoCase( DayWeekStr, Copy( S, 1, Length( DayWeekStr ) ) ) = 0 then
      begin
        Inc( S, Length( DayWeekStr ) );
        Exit;
      end;
      Dt := Dt + 1.0;
    end;
  end;

  procedure GetTimeMark( const fmt: KOLString; var S: PKOLChar );
  var SD: TSystemTime;
      AM: Boolean;
      C, TimeMarkStr: KOLString;
  begin
    GetSystemTime( SD );
    SD.wHour := 0;
    for AM := FALSE to TRUE do
    begin
      C := SystemDate2Str( SD, LOCALE_USER_DEFAULT, dfLongDate, PKOLChar( fmt + '/HH/mm' ) );
      TimeMarkStr := Parse( C, '/' );
      if AnsiCompareStrNoCase( TimeMarkStr, Copy( S, 1, Length( TimeMarkStr ) ) ) = 0 then
      begin
        Inc( S, Length( TimeMarkStr ) );
        hAM := AM;
        Exit;
      end;
      SD.wHour := 13;
    end;
    Result := 1;
  end;

  function FmtIs1( S: PKOLChar ): Boolean;
  begin
    if StrIsStartingFrom( FmtStr, S ) then
    begin
      Inc( FmtStr, {$IFDEF UNICODE_CTRLS} WStrLen {$ELSE} StrLen {$ENDIF}( S ) );
      Result := TRUE;
    end
      else
      Result := FALSE;
  end;

  function FmtIs( S1, S2: PKOLChar ): Boolean;
  begin
    Result := FmtIs1( S1 ) or FmtIs1( S2 );
  end;

  function SkipNoDig(const s:KOLString) :KOLString;
    var i:Integer;
  begin
    Result := s;
    i := IndexOfCharsMin(Result,'0123456789');
    if i>0 then
      Result := CopyEnd(Result,i);
  end;

var ST: TSystemTime;
begin
  sS:=SkipNoDig(sStr);
  FmtStr := PKOLChar( sFmtStr);
  S := PKOLChar( sS );
  FillChar( ST, Sizeof( ST ), #0 );
  h12 := FALSE;
  hAM := FALSE;
  while (FmtStr^ <> #0) and (S^ <> #0) do
  begin
    {$IFDEF UNICODE_CTRLS}
    if ((FmtStr^ >= 'a') and (FmtStr^ <= 'z') or
       (FmtStr^ >= 'A') and (FmtStr^ <= 'Z')) and
       (S^ >= '0') and (S^ <= '9') then
    {$ELSE}
    if (FmtStr^ in ['a'..'z','A'..'Z']) and (S^ in ['0'..'9']) then
    {$ENDIF}
    begin
      if      FmtIs1( 'yyyy'   ) then ST.wYear := GetNum( S, 4 )
      else if FmtIs1( 'yy' )     then ST.wYear := GetYear( S, 2 )
      else if FmtIs1( 'y' )      then ST.wYear := GetYear( S, -1 )
      else if FmtIs( 'dd', 'd' ) then ST.wDay := GetNum( S, 2 )
      else if FmtIs( 'MM', 'M' ) then ST.wMonth := GetNum( S, 2 )
      else if FmtIs( 'HH', 'H' ) then ST.wHour := GetNum( S, 2 )
      else if FmtIs( 'hh', 'h' ) then begin ST.wHour := GetNum( S, 2 ); h12 := TRUE end
      else if FmtIs( 'mm', 'm' ) then ST.wMinute := GetNum( S, 2 )
      else if FmtIs( 'ss', 's' ) then ST.wSecond := GetNum( S, 2 )
      else break; // + ECM
    end
      else
    {$IFDEF UNICODE_CTRLS}
    if (FmtStr^ = 'M') or (FmtStr^ = 'd') or (FmtStr^ = 'g') then
    {$ELSE}
    if (FmtStr^ in [ 'M', 'd', 'g' ]) then
    {$ENDIF}
    begin
      if      FmtIs1( 'MMMM' ) then ST.wMonth := GetMonth( 'MMMM', S )
      else if FmtIs1( 'MMM'  ) then ST.wMonth := GetMonth( 'MMM', S )
      else if FmtIs1( 'dddd' ) then SkipDayOfWeek( 'dddd', S )
      else if FmtIs1( 'ddd'  ) then SkipDayOfWeek( 'ddd', S )
      else if FmtIs1( 'tt'   ) then GetTimeMark( 'tt', S )
      else if FmtIs1( 't'    ) then GetTimeMark( 't', S )
      else break; // + ECM
    end
      else
    begin
      if FmtStr^ = S^ then
        Inc( FmtStr );
      Inc( S );
    end;
  end;

  if h12 then
  if hAM then
    Inc( ST.wHour, 12 );

  with ST do
    if (wHour=0) and (wMinute=0) and (wSecond=0) then begin
      wMinute:=Str2Int(SkipNoDig(S)) mod $10000;
      wHour := (wMinute div 60) mod 24;
      wMinute := wMinute mod 60;
    end;

  SystemTime2DateTime( ST, Result );
end;

var FmtBuf: PKOLChar;
    DateSeparator : KOLChar = #0; // + ECM

//[function Str2DateTimeShort]
function Str2DateTimeShort( const S: KOLString ): TDateTime;
var FmtStr, FmtStr2: KOLString;

  function EnumDateFmt( lpstrFmt: PKOLChar ): Boolean; stdcall;
  begin
    GetMem( FmtBuf, ({$IFDEF UNICODE_CTRLS} WStrLen {$ELSE} StrLen {$ENDIF}
      ( lpstrFmt ) + 1) * Sizeof( KOLChar ) );
    {$IFDEF UNICODE_CTRLS} WStrCopy {$ELSE} StrCopy {$ENDIF}
      ( FmtBuf, lpstrFmt );
    Result := FALSE;
  end;

begin
  FmtStr := 'dd.MM.yyyy';
  FmtBuf := nil;
  EnumDateFormats( @ EnumDateFmt, LOCALE_USER_DEFAULT, DATE_SHORTDATE );
  if FmtBuf <> nil then
  begin
    FmtStr := FmtBuf;
    FreeMem( FmtBuf );
  end;

  FmtStr2 := 'H:mm:ss';
  FmtBuf := nil;
  EnumTimeFormats( @ EnumDateFmt, LOCALE_USER_DEFAULT, 0 );
  if FmtBuf <> nil then
  begin
    FmtStr2 := FmtBuf;
    FreeMem( FmtBuf );
  end;

  Result := Str2DateTimeFmt( FmtStr + ' ' + FmtStr2, S );
end;

//[function Str2DateTimeShortEx]
function Str2DateTimeShortEx( const S: KOLString ): TDateTime;
var St: KOLString;
  Buff: Array[0..1] of KOLChar;
begin
  if DateSeparator = #0 then
  begin
    if GetLocaleInfo(GetThreadLocale,LOCALE_SDATE,Buff,2) > 0 then
      DateSeparator := Buff[0];
  end;
  St := S;
//  if Pos(DateSeparator,S) = 0 then    St := '0.0.0 '+S;
  Result := Str2DateTimeShort(St);
end;

function Str2FileTime(DateTimeStr: String; var FT: TFileTime) : Boolean;
  var SystemTime : TSystemTime;
begin
  Result := TRUE;
  if DateTimeStr = '' then
    GetLocalTime( SystemTime )  //GetSystemTime( SystemTime )//
  else begin
    Result :=
    DateTime2SystemTime(Str2DateTimeShortEx(DateTimeStr),SystemTime);
  end;
  if Result then
    Result :=
    SystemTimeToFileTime( SystemTime, FT );
  LocalFileTimeToFileTime(FT,FT);
end;


{ === See ======================================================
http://giscenter.icc.ru:8082/scripts/WWWBinV.dll/ShowR?DCU32.RFI
and the DCU32INT utility by Alexei Hmelnov (alex@monster.icc.ru)
http://monster.icc.ru/~alex/DCU/ }
type
  TVerD=( verUnknow,
    v1,
    verD2,verD3,verD4,verD5,verD6,verD7, verD8,verD9,verD10
    {,verK1,verK2,verK3});

function GetDcuVer(FName :string; var FTime :FileTime):TVerD;
var
  HFile: THandle;
  Magic, NmbRd :cardinal;
begin
  Result := verUnknow;
  if (FNAme = '') then EXIT;
  HFile := CreateFile(PChar(FNAme),
             {FILE_READ_ATTRIBUTES} GENERIC_READ, 0, nil,
             OPEN_EXISTING,
             FILE_ATTRIBUTE_NORMAL,
             0);
  if HFile = INVALID_HANDLE_VALUE then EXIT;

  if GetFileTime(HFile, nil, nil, @FTime) then begin
     if not ReadFile( HFile, Magic, SizeOf(Magic),NmbRd, nil ) then
       if SizeOf(Magic) <> NmbRd  then Magic := 0;
  end;
  CloseHandle(HFile);

  case Magic and ulong($FF80FFFF) of
    ulong($FF0000DF): Result := verD7;
    $0E0000DD,$0E8000DD: Result := verD6;
  else
    case Magic and ulong($FF00FFFF) of
      $0F0000DF: Result := {Free} verD7;
    else
      case Magic of
        $50505348: Result := verD2;
        $44518641: Result := verD3;
        $4768A6D8: Result := verD4;
        ulong($F21F148B): Result := verD5;
        $10000229: Result := verD8;
        $11000239, $1100000D, $11800009: Result := verD9;
        $12000023: Result := verD10; {not completely tested yet}
    {     ulong($F21F148C): Result := verK1; //Kylix 1.0
          $0E1011DD,$0E0001DD: Result := verK2; //Kylix 2.0
          $0F1001DD: Result := verK3; //Kylix 3.0}
      end ;{case}
    end ;{case}
  end ;{case}
end;

////////////////////////////////

function GetSwitches(const SwParam : String) :TSwitches;
  var i: Integer;
begin
  Result := [];
  for i := 2 to Length(SwParam) do
    case UpCase(SwParam[i]) of
    'W': Result := Result+[w];
    'A': Result := Result+[a];
    'C': Result := Result+[c];
    'F': Result := Result+[f];
    'D': Result := Result+[d];
    end;
end;


var
  FileNm, RefFileNm :string;
  Switches :TSwitches;
  FTime, FDcuTime :FileTime;
  HFile: THandle;
  lpFTimes:record
    lpCreationTime, lpLastAccessTime, lpLastWriteTime: PFileTime;
  end;
  RefFTimes:record
    CreationTime, LastAccessTime, LastWriteTime: FileTime;
  end;

  ok, RefFileErr : Boolean;
  ParamNr : Integer;

  VerD :TVerD =verUnknow;
  SysTime :TSystemTime;
const
  FILE_READ_ATTRIBUTES = ($0080); // all
  FILE_WRITE_ATTRIBUTES = ($0100); // all
BEGIN
  ok := FALSE;
  RefFileErr := FALSE;
  ParamNr := 1;
  FileNm:=ParamStr(ParamNr);
  if (FileNm <> '') then begin
    //Switches
    Switches := [w];
    inc(ParamNr);
    RefFileNm := ParamStr(ParamNr);
    if (Length(RefFileNm)>0) and (RefFileNm[1]='/') then begin
      if (Length(RefFileNm)=1)
      or  ( Length(RefFileNm)>1)
            and (Upcase(RefFileNm[2]) in ['W','A','C','F','D']) then begin
        Switches := GetSwitches(RefFileNm);
        if Switches=[] then Switches := [w];
        inc(ParamNr);
        if (f in Switches) then begin
          RefFileNm := ParamStr(ParamNr);
          inc(ParamNr);
        end;
      end;
    end;

    if (f in Switches) then begin
      RefFileErr := TRUE;
      if (RefFileNm <> '') then begin
        HFile := CreateFile(PChar(RefFileNm),
                   FILE_READ_ATTRIBUTES, 0, nil,
                   OPEN_EXISTING,
                                                //FILE_ATTRIBUTE_NORMAL,
                   FILE_FLAG_BACKUP_SEMANTICS, //get handle to directory or file

                   0);
        if HFile <> INVALID_HANDLE_VALUE then with RefFTimes do begin
          if GetFileTime(HFile, @CreationTime, @LastAccessTime, @LastWriteTime) then
            RefFileErr := FALSE;
          CloseHandle(HFile);
        end;
      end;
    end;

    if not RefFileErr then begin
      if (f in Switches)
      or ( ((Switches * [w,a,c]) <> [])
           and Str2FileTime( ParamStrEnd(ParamNr), FTime) )
      or ( (Switches * [d]) <> [] )
      then begin
        FillChar(lpFTimes,SizeOf(lpFTimes),0);
        if (f in Switches) then with lpFTimes, RefFTimes do begin
          lpLastWriteTime  := @LastWriteTime ;
          lpLastAccessTime := @LastAccessTime;
          lpCreationTime   := @CreationTime;
        end;
        with lpFTimes do begin
          if (w in Switches) then lpLastWriteTime  := @FTime;
          if (a in Switches) then lpLastAccessTime := @FTime;
          if (c in Switches) then lpCreationTime   := @FTime;
        end;

        if (d in Switches) then begin
          VerD := GetDcuVer(FileNm, FDcuTime);
          if VerD > VerUnknow then begin
            if lpFTimes.lpLastWriteTime <> nil then
              FDcuTime := lpFTimes.lpLastWriteTime^;
            //ustal nowe godziny i minuty
            FileTimeToSystemTime( FDcuTime, SysTime );
            SysTime.wHour:=ord(VerD);
            SysTime.wMinute:=0;
            SysTime.wSecond:=0;
            SysTime.wMilliseconds:=0;
            SystemTimeToFileTime(SysTime, FDcuTime);
            LocalFileTimeToFileTime(FDcuTime,FDcuTime);
            lpFTimes.lpLastWriteTime := @FDcuTime;
          end;
        end;

        //HFile:=FileCreate(FileNm, ofOpenWrite or ofOpenAlways);
        HFile := CreateFile(PChar(FileNm),
                   FILE_WRITE_ATTRIBUTES, 0, nil,
                   OPEN_EXISTING,
                                                //FILE_ATTRIBUTE_NORMAL,
                   FILE_FLAG_BACKUP_SEMANTICS, //get handle to directory or file
                   0);
        if HFile <> INVALID_HANDLE_VALUE then with lpFTimes do begin
          if SetFileTime(HFile,lpCreationTime, lpLastAccessTime, lpLastWriteTime)
          then begin
            ok := TRUE;
            FileTimeToLocalFileTime(lpLastWriteTime^,FTime);
            writeln(FileNm+' '+FileTime2Str(FTime));
          end;
          CloseHandle(HFile);//FileClose(HFile);
        end;{if HFile}
      end;
    end;
  end;

  if not ok then begin
    writeln(Info+DateTime2StrShort(now));
{$IFDEF PAUSE}
    write('Press ENTER...'); readln;
{$ENDIF PAUSE}
    HALT(1013);
  end
  else begin
{$IFDEF PAUSE}
    write('Press ENTER...'); readln;
{$ENDIF PAUSE}
    if (d in Switches) then begin
      if VerD = VerUnknow then HALT(1013)
      else HALT(ord(VerD));
    end
    else
    HALT(0);
  end;
END.
